@@ -33,8 +33,7 @@
 
 #include "libpq-fe.h"
 
-#include "watchdog/watchdog.h"
-#include "watchdog/wd_ext.h"
+#include "watchdog/wd_ipc_commands.h"
 
 #define WAIT_RETRY_COUNT (pool_config->recovery_timeout / 3)
 
@@ -49,7 +48,7 @@ static void check_postmaster_started(BackendInfo *backend);
 
 static char recovery_command[1024];
 
-extern volatile sig_atomic_t pcp_wakeup_request;
+extern volatile sig_atomic_t pcp_worker_wakeup_request;
 
 /*
  * Start online recovery.
@@ -72,10 +71,6 @@ void start_recovery(int recovery_node)
 		ereport(ERROR,
 				(errmsg("node recovery failed, node id: %d is not valid", recovery_node)));
 
-	if (*(my_backend_status[(recovery_node)]) == CON_UNUSED)
-		ereport(ERROR,
-				(errmsg("node recovery failed, node id: %d is unused", recovery_node)));
-
 	if (VALID_BACKEND(recovery_node))
 		ereport(ERROR,
 				(errmsg("node recovery failed, node id: %d is alive", recovery_node)));
@@ -117,7 +112,7 @@ void start_recovery(int recovery_node)
 			if (pool_config->use_watchdog)
 			{
 				/* announce start recovery */
-				if (WD_OK != wd_start_recovery())
+				if (COMMAND_OK != wd_start_recovery())
 					ereport(ERROR,
 							(errmsg("node recovery failed, failed to send start recovery packet")));
 			}
@@ -149,14 +144,14 @@ void start_recovery(int recovery_node)
 		 * reset failover completion flag.  this is necessary since
 		 * previous failover/failback will set the flag to 1.
 		 */
-		pcp_wakeup_request = 0;
+		pcp_worker_wakeup_request = 0;
 
 		/* send failback request to pgpool parent */
-		send_failback_request(recovery_node);
+		send_failback_request(recovery_node,false, false);
 
 		/* wait for failback */
 		failback_wait_count = 0;
-		while (!pcp_wakeup_request)
+		while (!pcp_worker_wakeup_request)
 		{
 			struct timeval t = {1, 0};
 			/* polling SIGUSR2 signal every 1 sec */
@@ -172,7 +167,7 @@ void start_recovery(int recovery_node)
 				break;
 			}
 		}
-		pcp_wakeup_request = 0;
+		pcp_worker_wakeup_request = 0;
 	}
 	PG_CATCH();
 	{
@@ -446,33 +441,8 @@ int wait_connection_closed(void)
 
 		if (WAIT_RETRY_COUNT != 0)
 			sleep(3);
-
 	} while (i++ < WAIT_RETRY_COUNT);
 	ereport(LOG,
 			(errmsg("wait_connection_closed: existing connections did not close in %d sec.", pool_config->recovery_timeout)));
-
-	/*
-	 * recovery_timeout was expired. Before returning with failure status,
-	 * let's check if this is caused by the malformed conn_counter. If a child
-	 * process abnormally exits (killed by SIGKILL or SEGFAULT, for example),
-	 * then conn_counter is not decremented at process exit, thus it will
-	 * never be returning to 0. This could be detected by checking if
-	 * client_idle_limit_in_recovery is enabled and less value than
-	 * recovery_timeout because all clients must be kicked out by the time
-	 * when client_idle_limit_in_recovery is expired. If so, we should reset
-	 * conn_counter to 0 also.
-	 *
-	 * See bug 431 for more info.
-	 */
-	if (pool_config->client_idle_limit_in_recovery == -1 ||
-		(pool_config->client_idle_limit_in_recovery > 0 &&
-		 pool_config->recovery_timeout >= pool_config->client_idle_limit_in_recovery))
-	{
-		ereport(LOG,
-				(errmsg("wait_connection_closed: mulformed conn_counter (%d) detected. reset it to 0",
-						Req_info->conn_counter)));
-		Req_info->conn_counter = 0;
-		return 0;
-	}
 	return 1;
 }
